@@ -1,5 +1,4 @@
-"""
-Implements streaming operations on table-like data and csv files.
+"""Implements streaming operations on table-like data and csv files.
 
 Conversions are defined in realtime based on table headers and called methods:
  - update
@@ -8,14 +7,28 @@ Conversions are defined in realtime based on table headers and called methods:
  - join
 
 """
-import csv
-import typing as t  # pylint: disable=unused-import
-from collections.abc import Sized
-from itertools import chain, zip_longest
 
-from ..base import (
+import csv
+from collections.abc import Callable, Sized
+from itertools import chain, zip_longest
+from typing import (
+    Any,
+    Iterable,
+    List,
+    Mapping,
+    Optional,
+    Sequence,
+    TextIO,
+    Tuple,
+    Union,
+    cast,
+)
+
+from .._aggregations import Aggregate, GroupBy, ReduceFuncs
+from .._base import (
     And,
     BaseConversion,
+    CallFunc,
     ConverterOptionsCtx,
     EscapedString,
     GeneratorComp,
@@ -25,40 +38,48 @@ from ..base import (
     InputArg,
     NaiveConversion,
     This,
-    Tuple,
+    Tuple_,
     ensure_conversion,
 )
-from ..columns import ColumnChanges, ColumnRef, MetaColumns
-from ..joins import JoinConversion, LeftJoinCondition, RightJoinCondition
+from .._columns import ColumnChanges, ColumnRef, MetaColumns
+from .._joins import JoinConversion, LeftJoinCondition, RightJoinCondition
 
 
 _none = BaseConversion._none
 
 
 class CloseFileIterator:
-    """
-    Utility to close the corresponding file once the iterator is exhausted
-    """
+    """Iterator wrapper, which closes file in the end."""
 
     def __init__(self, file_to_close):
+        """Init instance.
+
+        Args:
+          file_to_close: file descriptor to be closed the first time __next__
+            is called on this wrapper
+        """
         self.file_to_close = file_to_close
 
     def __iter__(self):
+        """Returns empty iterator."""
         return self
 
     def __next__(self):
+        """Stops iteration and closes the underlying file descriptor."""
         self.file_to_close.close()
         self.file_to_close = None
         raise StopIteration
 
     def __del__(self):
+        """Closes the file descriptor on garbage collection."""
+        # Consider rewriting with: weakref.finalize
         if self.file_to_close:
             self.file_to_close.close()
             self.file_to_close = None
 
 
 class CustomCsvDialect(csv.Dialect):
-    """A helper to define custom csv dialects without defining classes"""
+    """Create custom csv dialects without defining classes."""
 
     def __init__(
         self,
@@ -70,6 +91,7 @@ class CustomCsvDialect(csv.Dialect):
         lineterminator=csv.excel.lineterminator,
         quoting=csv.excel.quoting,
     ):
+        """Init self."""
         self.delimiter = delimiter
         self.quotechar = quotechar
         self.escapechar = escapechar
@@ -81,8 +103,7 @@ class CustomCsvDialect(csv.Dialect):
 
 
 class Table:
-    """Table conversion exposes streaming operations on table-like data and csv
-    files
+    """Streaming operations on table-like data and csv files.
 
     >>> Table.from_csv("input.csv", header=True).update(
     >>>     c=c.item("a") + c.item("b")
@@ -106,24 +127,24 @@ class Table:
 
     def __init__(
         self,
-        row_type: "t.Optional[type]",
-        rows_objects: "t.List[t.Iterable]",
+        row_type: "Optional[type]",
+        rows_objects: "List[Iterable]",
         meta_columns: "MetaColumns",
         pending_changes: int,
-        pipeline: "t.Optional[BaseConversion]" = None,
+        pipeline: "Optional[BaseConversion]" = None,
         file_to_close=None,
     ):
-        """It is used internally only. Use from_rows and from_csv methods."""
+        """For internal use only. Use from_rows and from_csv methods."""
         self.row_type = row_type
-        self.rows_objects: "t.Optional[t.List[t.Iterable]]" = rows_objects
+        self.rows_objects: "Optional[List[Iterable]]" = rows_objects
         self.meta_columns = meta_columns
         self.pending_changes = pending_changes
         self.pipeline = pipeline
         self.file_to_close = file_to_close
         self.row_type = row_type
 
-    def get_columns(self) -> "t.List[str]":
-        """Exposes list of column names"""
+    def get_columns(self) -> "List[str]":
+        """Return list of column names."""
         return [column.name for column in self.meta_columns.columns]
 
     columns = property(get_columns)
@@ -131,19 +152,18 @@ class Table:
     @classmethod
     def from_rows(
         cls,
-        rows: "t.Iterable[t.Union[dict, tuple, list, t.Any]]",
-        header: """t.Optional[
-            t.Union[
-                bool, t.List[str], t.Tuple[str], t.Dict[str, t.Union[str, int]]
+        rows: "Iterable[Union[dict, tuple, list, Any]]",
+        header: """Optional[
+            Union[
+                bool, List[str], Tuple[str, ...], Mapping[str, Union[str, int]]
             ]
         ]""" = None,
-        # t.Literal["raise", "keep", "drop", "mangle"]
+        # Literal["raise", "keep", "drop", "mangle"]
         duplicate_columns: str = "raise",
         skip_rows: int = 0,
         file_to_close=None,
     ) -> "Table":
-        """A method to initialize a table conversion from an iterable of
-        rows.
+        """Initialize a table from an iterable of rows.
 
         Args:
           rows: can be either an iterable of any objects if no header inference
@@ -176,6 +196,8 @@ class Table:
             data contains a header, but you provide your own - in this case
             it's convenient to skip the heading row from the input
 
+          file_to_close: for internal use
+
         """
         columns = MetaColumns(duplicate_columns=duplicate_columns)
 
@@ -185,14 +207,19 @@ class Table:
             for _ in range(skip_rows):
                 next(rows)
 
-        first_row = next(rows)
+        first_row = next(rows, _none)
         first_row_is_sized = isinstance(first_row, Sized)
-        row_type = type(first_row)
+        row_type = tuple if first_row is _none else type(first_row)
         pending_changes = 0
 
-        index: "t.Union[str, int]"
+        index: "Union[str, int]"
         if isinstance(header, (tuple, list)):
-            if first_row_is_sized:
+            if first_row is _none:
+                for index, name in enumerate(header):
+                    pending_changes |= columns.add(name, index, None)[1]
+
+            elif first_row_is_sized:
+                first_row = cast(Sized, first_row)
                 if len(header) != len(first_row):
                     raise ValueError("non-matching number of columns")
                 for index, column in enumerate(header):
@@ -203,48 +230,57 @@ class Table:
                 pending_changes |= columns.add(header[0], None, This())[1]
 
         elif isinstance(header, dict):
-            if not first_row_is_sized:
+            if first_row is _none:
+                for name, index in header.items():
+                    pending_changes |= columns.add(name, index, None)[1]
+
+            elif first_row_is_sized:
+                first_row = cast(Sized, first_row)
+                if len(header) != len(first_row):
+                    raise ValueError("non-matching number of columns")
+                for name, index in header.items():
+                    pending_changes |= columns.add(name, index, None)[1]
+            else:
                 raise ValueError(
                     "first row is not sized to apply header as dict"
                 )
-            if len(header) != len(first_row):
-                raise ValueError("non-matching number of columns")
-            for name, index in header.items():
-                pending_changes |= columns.add(name, index, None)[1]
 
-        else:
-            # inferring a header
-            if isinstance(first_row, dict):
-                if header is False:
-                    for key in first_row:
-                        pending_changes |= columns.add(None, key, None)[1]
-                    pending_changes |= ColumnChanges.MUTATE
-                else:
-                    for key in first_row:
-                        pending_changes |= columns.add(key, key, None)[1]
+        # inferring a header
+        elif isinstance(first_row, dict):
+            if header is False:
+                for key in first_row:
+                    pending_changes |= columns.add(None, key, None)[1]
+                pending_changes |= ColumnChanges.MUTATE
+            else:
+                for key in first_row:
+                    pending_changes |= columns.add(key, key, None)[1]
 
-            elif isinstance(first_row, (tuple, list)):
-                if header is True:
-                    for index, column_name in enumerate(first_row):
-                        pending_changes |= columns.add(
-                            column_name, index, None
-                        )[1]
-                    first_row = None
-
-                else:
-                    for index in range(len(first_row)):
-                        pending_changes |= columns.add(None, index, None)[1]
+        elif isinstance(first_row, (tuple, list)):
+            if header is True:
+                for index, column_name in enumerate(first_row):
+                    pending_changes |= columns.add(column_name, index, None)[1]
+                first_row = None
 
             else:
-                if header is True:
+                for index in range(len(first_row)):
+                    pending_changes |= columns.add(None, index, None)[1]
+
+        else:
+            if header is True:
+                if first_row is _none:
+                    raise ValueError(
+                        "impossible to infer header of an empty sequence"
+                    )
+                else:
                     pending_changes |= columns.add(first_row, None, This())[1]
                     first_row = None
-                else:
-                    pending_changes |= columns.add(None, None, This())[1]
-                pending_changes |= ColumnChanges.MUTATE
+            else:
+                pending_changes |= columns.add(None, None, This())[1]
 
-        rows_objects: "t.List[t.Iterable]" = [rows]
-        if first_row is not None:
+            pending_changes |= ColumnChanges.MUTATE
+
+        rows_objects: "List[Iterable]" = [rows]
+        if first_row is not None and first_row is not _none:
             rows_objects.insert(0, (first_row,))
 
         return cls(
@@ -260,24 +296,24 @@ class Table:
     @classmethod
     def from_csv(
         cls,
-        filepath_or_buffer: "t.Union[str, t.TextIO]",
-        header: """t.Optional[
-            t.Union[
-                bool, t.List[str], t.Tuple[str], t.Dict[str, t.Union[str, int]]
+        filepath_or_buffer: "Union[str, TextIO]",
+        header: """Optional[
+            Union[
+                bool, List[str], Tuple[str], Mapping[str, Union[str, int]]
             ]
         ]""" = None,
         duplicate_columns: str = "mangle",
         skip_rows: int = 0,
-        dialect: "t.Union[str, CustomCsvDialect]" = "excel",
+        dialect: "Union[str, CustomCsvDialect]" = "excel",
         encoding: str = "utf-8",
     ) -> "Table":
-        """A method to initialize a table conversion from a csv-like file.
+        """Initialize a table conversion from a csv-like object.
 
         Args:
           filepath_or_buffer: a filepath or something :py:obj:`csv.reader` can
             read
-          header: specifies header inference mode:
 
+          header: specifies header inference mode:
              * True: takes either the first tuple/list or keys of the first
                dict as a header
              * False: there's no header in input data, use numbered columns
@@ -292,7 +328,6 @@ class Table:
                keys as a header
 
           duplicate_columns: either of following ("mangle" by default):
-
             * "raise": ValueError is raise if a duplicate column is detected
             * "keep": duplicate columns are left as is, but when referenced the
               first one is used
@@ -308,18 +343,18 @@ class Table:
             helper method:
             :py:obj:`convtools.contrib.tables.Table.csv_dialect` to create
             dialects without defining classes
+
           encoding: encoding to pass to :py:obj:`open`
         """
-
-        file_to_close: "t.Optional[t.TextIO]"
-        buffer: "t.TextIO"
+        file_to_close: "Optional[TextIO]"
+        buffer: "TextIO"
         if isinstance(filepath_or_buffer, str):
-            buffer = (
-                file_to_close
-            ) = open(  # pylint: disable=consider-using-with
-                filepath_or_buffer,
-                "r",
-                encoding=encoding,
+            buffer = file_to_close = (
+                open(  # pylint: disable=consider-using-with # noqa: SIM115
+                    filepath_or_buffer,
+                    "r",
+                    encoding=encoding,
+                )
             )
         else:
             buffer = filepath_or_buffer
@@ -339,7 +374,9 @@ class Table:
         )
 
     def embed_conversions(self) -> "Table":
-        """There's no need in calling this directly, it's done automatically
+        """For internal use only.
+
+        There's no need in calling this directly, it's done automatically
         when you use :py:obj:`convtools.contrib.tables.Table.update` or
         :py:obj:`convtools.contrib.tables.Table.join` See the explanation
         below:
@@ -377,24 +414,22 @@ class Table:
 
         return self
 
-    def filter(self, condition: "BaseConversion") -> "Table":
-        """Filters table-like data, keeping rows where ``condition`` resolves
-        to ``True``"""
-        condition = ensure_conversion(condition)
-        column_refs = list(condition.get_dependencies(types=ColumnRef))
-        name_to_column = self.meta_columns.get_name_to_column()
+    def _set_col_indexes(self, name_to_column, conversion):
+        conversion = ensure_conversion(conversion)
 
-        depends_on_complex_columns = any(
-            name_to_column[ref.name].conversion is not None
-            for ref in column_refs
-        )
-        if depends_on_complex_columns:
-            return self.embed_conversions().filter(condition)
-
-        for ref in condition.get_dependencies(types=ColumnRef):
+        for ref in conversion.get_dependencies(types=ColumnRef):
             column = name_to_column[ref.name]
+            if column.conversion is not None:
+                self.embed_conversions()
             ref.set_index(column.index)
 
+        return conversion
+
+    def filter(self, condition: "BaseConversion") -> "Table":
+        """Keep rows where ``condition`` resolves to `True`."""
+        condition = self._set_col_indexes(
+            self.meta_columns.get_name_to_column(), condition
+        )
         self.pipeline = (self.pipeline or This()).filter(condition)
         return self
 
@@ -408,20 +443,9 @@ class Table:
         column_name_to_column = self.meta_columns.get_name_to_column()
 
         for column_name in list(column_to_conversion):
-            conversion = ensure_conversion(column_to_conversion[column_name])
-            column_refs = list(conversion.get_dependencies(types=ColumnRef))
-
-            depends_on_complex_columns = any(
-                column_name_to_column[ref.name].conversion is not None
-                for ref in column_refs
+            conversion = self._set_col_indexes(
+                column_name_to_column, column_to_conversion[column_name]
             )
-            if depends_on_complex_columns:
-                return self.embed_conversions().update(**column_to_conversion)
-
-            del column_to_conversion[column_name]
-
-            for ref in column_refs:
-                ref.set_index(column_name_to_column[ref.name].index)
 
             if column_name in column_name_to_column:
                 column = column_name_to_column[column_name]
@@ -454,15 +478,17 @@ class Table:
         return self.update(**column_to_conversion)
 
     def rename(
-        self, columns: "t.Union[t.Tuple[str], t.List[str], t.Dict[str, str]]"
+        self, columns: "Union[Tuple[str], List[str], Mapping[str, str]]"
     ) -> "Table":
-        """Method to rename columns. The behavior depends on type of columns
-        argument.
+        """Rename columns.
+
+        The behavior depends on type of columns argument.
 
         Args:
-          columns: if tuple/list, then it defines new column names (length of
-            passed columns should match number of columns inside).  If dict,
-            then it defines a mapping from old column names to new ones.
+          columns:
+            * tuple/list: then it defines new column names (length of
+              passed columns should match number of columns inside).
+            * dict: defines a mapping from old column names to new ones.
         """
         renamed = False
         if isinstance(columns, dict):
@@ -488,9 +514,11 @@ class Table:
 
         return self
 
-    def take(self, *column_names: "t.Union[str, t.Any]") -> "Table":
-        """Leaves only specified columns, omitting the rest. ``...`` references
-        non-mentioned columns, so it's easy to both take a few columns:
+    def take(self, *column_names: "Union[str, Any]") -> "Table":
+        """Leave only specified columns, omitting the rest.
+
+        ``...`` references non-mentioned columns, so it's easy to both take a
+        few columns:
 
         >>> table.take("a", "b")
 
@@ -517,8 +545,10 @@ class Table:
         return self
 
     def zip(self, table: "Table", fill_value=None) -> "Table":
-        """Zip tables one to another. Before using this method, make sure you
-        are not looking for :py:obj:`convtools.contrib.tables.Table.join`
+        """Zip tables one to another.
+
+        Before using this method, make sure you are not looking for
+        `convtools.contrib.tables.Table.join`
 
         Let's assume fill_value is set to " ":
 
@@ -535,9 +565,8 @@ class Table:
         >>> |   |   | 5 | 6 |
 
         Args:
-         - table: table to be chained
-         - fill_value: value to use for filling gaps
-
+          table: table to be chained
+          fill_value: value to use for filling gaps
         """
         new_columns = MetaColumns(duplicate_columns="keep")
         left_columns = self.meta_columns.get_name_to_column()
@@ -552,9 +581,11 @@ class Table:
             new_columns.add(name, None, GetItem(1, index))
 
         new_rows = (
-            (left_fill_value, t[1])
-            if t[0] is None
-            else ((t[0], right_fill_value) if t[1] is None else t)
+            (
+                (left_fill_value, t[1])
+                if t[0] is None
+                else ((t[0], right_fill_value) if t[1] is None else t)
+            )
             for t in zip_longest(
                 self.into_iter_rows(tuple), table.into_iter_rows(tuple)
             )
@@ -587,9 +618,8 @@ class Table:
         >>> |   | 3 | 4 |
 
         Args:
-         - table: table to be chained
-         - fill_value: value to use for filling gaps
-
+          table: table to be chained
+          fill_value: value to use for filling gaps
         """
         if self.rows_objects is None:
             raise AssertionError("move_rows called the 2nd time")
@@ -655,11 +685,11 @@ class Table:
     def join(
         self,
         table: "Table",
-        on: "t.Union[BaseConversion, str, t.Iterable[str]]",
+        on: "Union[BaseConversion, str, Iterable[str]]",
         how: str,
         suffixes=("_LEFT", "_RIGHT"),
     ) -> "Table":
-        """Joins the table conversion to another table conversion.
+        """Classic table join.
 
         Args:
           table: another table conversion to join to self
@@ -675,17 +705,25 @@ class Table:
             the second one is added to conflicting right ones. When ``on`` is
             an iterable of strings, these columns are excluded from suffixing.
         """
-
         how = JoinConversion.validate_how(how)
         left = self.embed_conversions()
         right = table.embed_conversions()
+
+        for table_ in (left, right):
+            # after embedding there are no conversions, only indexes.
+            # indexes can be used as is except for dict-based data after
+            # renamings, but if we skip rebuilding it from scratch by
+            # into_list_of_iterables we can continue using indexes
+            if table_.row_type is dict and table_.pending_changes:
+                table_.pending_changes = 0
+
         left_join_conversion = LeftJoinCondition()
         right_join_conversion = RightJoinCondition()
         left_column_name_to_column = left.meta_columns.get_name_to_column()
         right_column_name_to_column = right.meta_columns.get_name_to_column()
 
-        after_join_conversions: "t.List[BaseConversion]" = []
-        after_join_column_names: "t.List[str]" = []
+        after_join_conversions: "List[BaseConversion]" = []
+        after_join_column_names: "List[str]" = []
 
         if isinstance(on, BaseConversion):
             # intentionally left blank to force suffixing
@@ -803,8 +841,7 @@ class Table:
         )
 
     def explode(self, column_name: str):
-        """Explodes a table to a long format by exploding a column with
-        iterables, e.g.:
+        """Explode a table by unnesting a column with iterables.
 
         >>> | a |   b    |
         >>> | 1 | [2, 3] |
@@ -819,8 +856,7 @@ class Table:
         >>> | 2 | 6 |
 
         Args:
-         - column_name: column with iterables to be exploded
-
+          column_name: column with iterables to be exploded
         """
         columns = self.columns
         try:
@@ -837,7 +873,7 @@ class Table:
                 """({new_row} for {row} in {rows} for {value} in {values})"""
             )
             .pass_args(
-                new_row=Tuple(
+                new_row=Tuple_(
                     *(
                         c_value if i == index else c_row.item(i)
                         for i in range(len(columns))
@@ -853,8 +889,196 @@ class Table:
 
         return Table.from_rows(new_rows, header=columns)
 
-    def move_rows_objects(self) -> "t.List[t.Iterable]":
-        """Moves out rows objects including files to be closed later"""
+    def wide_to_long(
+        self,
+        col_for_names: str = "name",
+        col_for_values: str = "value",
+        prepare_name: "Optional[Callable[[str], str]]" = None,
+        prepare_value: "Optional[Callable[[str], str]]" = None,
+        keep_cols: "Sequence[str]" = (),
+    ):
+        """Turn wide table into a long form.
+
+        This is an operation, which is opposite to pivot operation:
+        >>> assert list(
+        >>>     Table.from_rows(
+        >>>         [{"name": "John", "height": 200, "age": 30, "mood": "good"}]
+        >>>     )
+        >>>     .wide_to_long(
+        >>>         col_for_names="metric",
+        >>>         col_for_values="value",
+        >>>         keep_cols=("name",)
+        >>>     )
+        >>>     .into_iter_rows(dict)
+        >>> ) == [
+        >>>     {"name": "John", "metric": "height", "value": 200},
+        >>>     {"name": "John", "metric": "age", "value": 30},
+        >>>     {"name": "John", "metric": "mood", "value": "good"},
+        >>> ]
+
+        Args:
+          col_for_names: name of the column with names of processed columns
+          col_for_values: name of the column with values of processed columns
+          prepare_name (optional): callable or conversion to prepare a name
+          prepare_value (optional): callable or conversion to prepare a value
+          keep_cols: column names to keep as is
+        Returns: a new Table.
+        """
+        columns = self.columns
+        keep_cols_set = set(keep_cols)
+        collapse_cols = tuple(
+            col for col in columns if col not in keep_cols_set
+        )
+        resulting_cols = tuple(keep_cols) + (col_for_names, col_for_values)
+
+        new_rows = (
+            InlineExpr(
+                """(new_row for row_ in {rows} for new_row in {new_rows})"""
+            )
+            .pass_args(
+                rows=This,
+                new_rows=Tuple_(
+                    *(
+                        (
+                            *(
+                                EscapedString("row_").item(
+                                    columns.index(keep_col)
+                                )
+                                for keep_col in keep_cols
+                            ),
+                            (
+                                collapse_col
+                                if prepare_name is None
+                                else This.pipe(prepare_name).execute(
+                                    collapse_col
+                                )
+                            ),
+                            EscapedString("row_")
+                            .item(columns.index(collapse_col))
+                            .pipe(
+                                This
+                                if prepare_value is None
+                                else prepare_value
+                            ),
+                        )
+                        for collapse_col in collapse_cols
+                    )
+                ),
+            )
+            .execute(self.into_iter_rows(tuple, include_header=False))
+        )
+        return Table.from_rows(new_rows, header=resulting_cols)
+
+    def pivot(
+        self,
+        rows: "Sequence[str]",
+        columns: "Sequence[str]",
+        values: "Mapping[str, BaseConversion]",
+        prepare_column_names: "Callable[[Sequence[str]], str]" = " - ".join,
+    ) -> "Table":
+        """Create a pivot table.
+
+        >>> assert list(
+        >>>     Table.from_rows([
+        >>>        {"dept": 1, "year": 2023, "currency": "USD", "revenue": 100},
+        >>>        {"dept": 1, "year": 2024, "currency": "USD", "revenue": 300},
+        >>>        {"dept": 1, "year": 2024, "currency": "CNY", "revenue": 200},
+        >>>        {"dept": 1, "year": 2024, "currency": "CNY", "revenue": 111},
+        >>>     ])
+        >>>     .pivot(
+        >>>         rows=["year", "dept"],
+        >>>         columns=["currency"],
+        >>>         values={
+        >>>             "sum": c.ReduceFuncs.Sum(c.col("revenue")),
+        >>>             "min": c.ReduceFuncs.Min(c.col("revenue")),
+        >>>         },
+        >>>     )
+        >>>     .into_iter_rows(dict)
+        >>> ) == [
+        >>>     {
+        >>>         "CNY - min": None,
+        >>>         "CNY - sum": None,
+        >>>         "USD - min": 100,
+        >>>         "USD - sum": 100,
+        >>>         "dept": 1,
+        >>>         "year": 2023,
+        >>>     },
+        >>>     {
+        >>>         "CNY - min": 111,
+        >>>         "CNY - sum": 311,
+        >>>         "USD - min": 300,
+        >>>         "USD - sum": 300,
+        >>>         "dept": 1,
+        >>>         "year": 2024,
+        >>>     },
+        >>> ]
+
+        Args:
+          rows: columns to group by
+          columns: columns to take names of new columns from
+          values: mapping of name to reducer of column value/values
+          prepare_column_names: callable to create column names from column
+            names and reducer name
+
+        """
+        name_to_column = self.meta_columns.get_name_to_column()
+        name_to_index_cols = {
+            col_name: self._set_col_indexes(
+                name_to_column, ColumnRef(col_name)
+            )
+            for col_name in rows
+        }
+        columns = [
+            self._set_col_indexes(name_to_column, ColumnRef(col_name))
+            for col_name in columns
+        ]
+        value_name_to_aggregator = {
+            key: Aggregate(
+                self._set_col_indexes(name_to_column, value)
+            ).gen_converter()
+            for key, value in values.items()
+        }
+
+        aggregated_data = (
+            GroupBy(*name_to_index_cols.values())
+            .aggregate(
+                (
+                    *name_to_index_cols.values(),
+                    ReduceFuncs.DictArray(tuple(columns), This()),
+                )
+            )
+            .execute(self.into_iter_rows(self.row_type))
+        )
+        index_ = len(name_to_index_cols)
+        new_col_keys = list(
+            {key: 0 for row in aggregated_data for key in row[index_]}
+        )
+
+        agg_data_col = "7b51817c-c2d2-4f9c-a708-65d776ed2ddd"
+        return (
+            Table.from_rows(
+                aggregated_data,
+                header=tuple(name_to_index_cols) + (agg_data_col,),
+            )
+            .update(
+                **{
+                    prepare_column_names(
+                        new_col_key + (value_name,)
+                    ): ColumnRef(agg_data_col)
+                    .call_method("get", NaiveConversion(new_col_key))
+                    .and_then(CallFunc(aggregator, This()))
+                    for new_col_key in new_col_keys
+                    for value_name, aggregator in value_name_to_aggregator.items()
+                }
+            )
+            .drop(agg_data_col)
+        )
+
+    def move_rows_objects(self) -> "List[Iterable]":
+        """For internal use.
+
+        Moves out rows objects including files to be closed later.
+        """
         if self.rows_objects is None:
             raise AssertionError("move_rows called the 2nd time")
 
@@ -870,7 +1094,8 @@ class Table:
 
     def into_list_of_iterables(
         self, type_=tuple, include_header=None
-    ) -> "t.List[t.Iterable]":
+    ) -> "List[Iterable]":
+        """For internal use."""
         if type_ not in self.supported_types:
             raise TypeError("unsupported type_", type_)
 
@@ -881,7 +1106,7 @@ class Table:
         if no_pending_changes:
             conversion = None if self.pipeline is None else self.pipeline
         else:
-            row_conversion: "t.Union[dict, tuple, list]"
+            row_conversion: "Union[dict, tuple, list]"
             if type_ is dict:
                 row_conversion = {
                     column.name: (
@@ -919,8 +1144,8 @@ class Table:
 
     def into_iter_rows(
         self, type_=tuple, include_header=None
-    ) -> "t.Iterable[t.Any]":
-        """Consumes inner rows Iterable and returns Iterable of processed rows.
+    ) -> "Iterable[Any]":
+        """Return iterable of processed rows.
 
         Args:
           type_: casts output rows to the type. Accepts the following values:
@@ -928,6 +1153,7 @@ class Table:
             * :py:obj:`dict`
             * :py:obj:`tuple`
             * :py:obj:`list`
+          include_header: whether to include header row in the result or not
 
         """
         return chain.from_iterable(
@@ -938,13 +1164,12 @@ class Table:
 
     def into_csv(
         self,
-        filepath_or_buffer: "t.Union[str, t.TextIO]",
+        filepath_or_buffer: "Union[str, TextIO]",
         include_header: bool = True,
-        dialect: "t.Union[str, CustomCsvDialect]" = "excel",
+        dialect: "Union[str, CustomCsvDialect]" = "excel",
         encoding="utf-8",
     ):
-        """Consumes inner rows Iterable and writes processed rows as a csv-like
-        file.
+        """Writes rows to a file.
 
         Args:
           filepath_or_buffer: a filepath or something :py:obj:`csv.writer` can
@@ -959,10 +1184,12 @@ class Table:
         row_type = list if self.row_type is list else tuple
 
         f_to_close = None
-        f: "t.TextIO"
+        f: "TextIO"
         if isinstance(filepath_or_buffer, str):
-            f = f_to_close = open(  # pylint:disable=consider-using-with
-                filepath_or_buffer, "w", encoding=encoding
+            f = f_to_close = (
+                open(  # pylint:disable=consider-using-with  # noqa: SIM115
+                    filepath_or_buffer, "w", encoding=encoding
+                )
             )
         else:
             f = filepath_or_buffer
